(** PCP Solver *)
module PCPParser = struct

open Parser

let int = many digit

let int_list =
 between 
  (char '[' >> spaces) 
  (sep_by (spaces >> char ';' >> spaces) int)
  (spaces >> char ']')

let parse = 
 spaces >>
 int_list >>= fun d1 ->
 spaces >>
 int_list >>= fun d2 ->
 spaces >>
 return (Lst.zip d1 d2)

let parse input = match Parser.parse parse (Strng.of_string input) with
 | None ->  
  Format.printf "shape of dictionaries: \"[0;01;110] [100;00;11]\"\n";
  failwith ("parsing error on " ^ input)
 | Some ds -> ds
end

(*samples*)
(*shortest solution [0;1]*)
let pcp1 = PCPParser.parse "[01;0] [0;10]";; 

(*no solution*)
let pcp2 = PCPParser.parse "[01;11] [0;10]";;

(*shortest solution of length 76*)
let pcp3 = PCPParser.parse "[01;01;0;100] [0;110;100;10]";;

(*solution of length 132*)
let pcp4 = PCPParser.parse "[101;0;10;11] [1;01;110;10]";;

(*no solution*)
let pcp5 = PCPParser.parse "[01;00] [0;10]";;

(*solving*)
(* trim : ('a list * 'a list) -> ('a list * 'a list) *)
let rec trim = function
 | (x::xs,y::ys) when x = y -> trim (xs,ys)
 | d -> d

(* extend : ('a list * 'a list) -> ('a list * 'a list) *)
let extend (w1,w2) (d1,d2) = trim (w1@d1,w2@d2)

(* solve : ('a list * 'a list) list -> bool *)
let solve ds =
 let rec solve = function
  | [] -> false
  | ([],[])::_ -> true
  | (x::_,y::_)::ws when x <> y -> solve ws
  | w::ws -> solve (ws@Lst.map (extend w) ds)
 in solve (Lst.map trim ds)

(* E13 *)
(* solve_dp : ('a list * 'a list) list -> bool *)
let solve_dp ds =
 let rec solve t = function
  | [] -> false
  | ([],[])::_ -> true
  | (x::_,y::_)::ws when x <> y -> solve t ws
  | w::ws -> if Lookup.mem t w then solve t ws else 
    let t = Lookup.add t w () in
    solve t (ws@(Lst.map (extend w) ds))
 in solve Lookup.empty (Lst.map trim ds)

(* extendi : int list -> ('a list * 'a list) -> int -> ('a list * 'a list) *)
let extendi is w i d = (i::is,extend w d)

(* solve_dps : ('a list * 'a list) list -> int list option *)
let solve_dps ds =
 let rec solve t = function
  | [] -> None
  | (is,([],[]))::_ -> Some (Lst.rev is)
  | (_,(x::_,y::_))::ws when x <> y -> solve t ws
  | (is,w)::ws -> if Lookup.mem t w then solve t ws else 
    let t = Lookup.add t w () in
    (* heuristic to cut off search *)
    (* if (abs (Lst.length (fst w) - Lst.length (snd w)) > 20) then solve t ws else *)
    solve t (ws@(Lst.mapi (extendi is w) ds))
 in solve Lookup.empty (Lst.mapi (extendi [] ([],[])) (Lst.map trim ds))

(**solver with priority queue**)
let delta (_,(p,q)) = abs (Lst.length p - Lst.length q)

let rec insert ws elt = match ws with
 | [] -> [elt]
 | w::ws -> if delta elt < delta w then elt::w::ws else w::insert ws elt 

(* solve_dpsq : ('a list * 'a list) list -> int list option *)
let solve_dpsq ds =
 let rec solve t = function
  | [] -> None
  | (is,([],[]))::_ -> Some (Lst.rev is)
  | (_,(x::_,y::_))::ws when x <> y -> solve t ws
  | (is,w)::ws -> if Lookup.mem t w then solve t ws else 
    let t = Lookup.add t w () in
    solve t (Lst.foldl insert ws (Lst.mapi (extendi is w) ds))
 in solve Lookup.empty (Lst.mapi (extendi [] ([],[])) (Lst.map trim ds))

(* pcp solver with parser *)
(*depth first search; finds shortest solution*)
let pcp d1 d2 = solve_dps (PCPParser.parse (d1 ^ " " ^ d2))

(*priority queue; solution need not be shortest*)
let pcpp d1 d2 = solve_dpsq (PCPParser.parse (d1 ^ " " ^ d2))
